// Copyright (c) 2025 Charlie Vanaret
// Licensed under the MIT license. See LICENSE file in the project directory for details.

#ifndef UNO_INDEXING_H
#define UNO_INDEXING_H

namespace uno {
   struct Indexing {
      static constexpr int C_indexing = 0;
      static constexpr int Fortran_indexing = 1;
   };
} // namespace

#endif // UNO_INDEXING_H