// Copyright (c) 2025 Charlie Vanaret
// Licensed under the MIT license. See LICENSE file in the project directory for details.

#ifndef UNO_UNOINT_H
#define UNO_UNOINT_H

#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef int32_t uno_int;

#ifdef __cplusplus
}
#endif

#endif // UNO_UNOINT_H