// Copyright (c) 2025 Charlie Vanaret
// Licensed under the MIT license. See LICENSE file in the project directory for details.

#ifndef UNO_MATRIXMAJOR_H
#define UNO_MATRIXMAJOR_H

namespace uno {
   enum class MatrixOrder {
      ROW_MAJOR,
      COLUMN_MAJOR
   };
} // namespace

#endif // UNO_MATRIXMAJOR_H